import 'package:bloc/bloc.dart';
import 'package:meta/meta.dart' show visibleForOverriding;
import '../bloc_extras_base.dart';
import 'list_adapter.dart';
import 'package:built_collection/src/list.dart';
import 'simple_list_state.dart';

part 'simple_list_event.dart';

abstract class SimpleListBlocBase<TItem>
    extends Bloc<_SimpleListEvent<TItem>, SimpleListState<TItem>> {
  SimpleListBlocBase() : super(SimpleListState.initial());

  @override
  Stream<SimpleListState<TItem>> mapEventToState(
    _SimpleListEvent<TItem> event,
  ) async* {
    if (state.loadingState.status == LoadingStatus.initial &&
        !(event is LoadItemsEvent<TItem>)) {
      throw StateError(
        'The items were not loaded yet.',
      );
    }

    if (event is _SingleItemEvent<TItem> &&
        !state.items.contains(event.item) &&
        !(event is CreateItemEvent<TItem>)) {
      throw StateError(
        'The item you try to act on does not exist in the list.',
      );
    }

    try {
      yield SimpleListState.loading(state.items);
      BuiltList<TItem> items;
      if (event is LoadItemsEvent<TItem>) {
        items = await _listAll();
      } else if (event is CreateItemEvent<TItem>) {
        items = state.items.append(await _create(event.item));
      } else if (event is DeleteItemEvent<TItem>) {
        await _delete(event.item);
        items = state.items.without(event.item);
      } else if (event is UpdateItemEvent<TItem>) {
        items = state.items.replaceItem(
          event.item,
          await _update(event.item, event.updates),
        );
      } else if (event is ReloadItemEvent<TItem>) {
        items = state.items.replaceItem(event.item, await _reload(event.item));
      }

      yield SimpleListState.ready(items ?? state.items);
    } catch (exception, stackTrace) {
      yield SimpleListState.error(
        error: exception.toString(),
        details: stackTrace.toString(),
        items: state.items,
      );
    }
  }

  @visibleForOverriding
  Future<BuiltList<TItem>> _listAll();

  @visibleForOverriding
  Future<TItem> _create(TItem item);

  @visibleForOverriding
  Future<void> _delete(TItem item);

  @visibleForOverriding
  Future<TItem> _update(TItem item, TItem updates);

  @visibleForOverriding
  Future<TItem> _reload(TItem item);
}

class SimpleListBloc<TItem> extends SimpleListBlocBase<TItem> {
  final ListAdapter<TItem> adapter;

  SimpleListBloc(this.adapter)
      : assert(adapter != null),
        super();

  @override
  Future<TItem> _reload(TItem item) async => await adapter.retrieve(item);

  @override
  Future<void> _delete(TItem item) => adapter.delete(item);

  @override
  Future<TItem> _create(TItem item) => adapter.create(item);

  @override
  Future<BuiltList<TItem>> _listAll() => adapter.listAll();

  @override
  Future<TItem> _update(TItem item, TItem updates) => adapter.update(updates);
}
