import 'package:bloc_test/bloc_test.dart';
import 'package:built_collection/built_collection.dart';
import 'package:built_value/built_value.dart';
import 'package:meta/meta.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';
import 'package:meta/meta.dart' show isTest;

import 'package:bloc_extras/bloc_extras.dart';

part 'simple_list_bloc_test.g.dart';

void main() {
  group('Simple list bloc', () {
    final adapter = _TestItemAdapterMock();
    final items = List<_TestItem>.generate(
        5, (index) => _TestItem(name: 'Item ${index + 1}', id: index + 1));

    tearDown(() {
      reset(adapter);
    });

    _blocTest(
      'Showld load all items',
      builder: () {
        when(adapter.listAll()).thenAnswer((realInvocation) async {
          return BuiltList(items);
        });
        return SimpleListBloc(adapter);
      },
      act: (cubit) => cubit.add(
        LoadItemsEvent(),
      ),
      expect: [
        SimpleListState<_TestItem>.loading(),
        SimpleListState<_TestItem>.ready(items),
      ],
      verify: (bloc) => verify(adapter.listAll()).called(1),
      wait: Duration.zero,
    );

    _blocTest(
      'Showld throw an error if the state is initial',
      builder: () => SimpleListBloc(adapter),
      act: (cubit) => cubit.add(CreateItemEvent(
        _TestItem(name: 'Named item'),
      )),
      errors: [
        isA<StateError>(),
      ],
    );

    _blocTest(
      'Showld remove the item',
      builder: () {
        when(adapter.delete(items[1]))
            .thenAnswer((realInvocation) => Future.delayed(Duration.zero));
        return SimpleListBloc((adapter));
      },
      seed: SimpleListState.ready(items),
      act: (cubit) => cubit.add(
        DeleteItemEvent(items[1]),
      ),
      expect: [
        SimpleListState.loading(items),
        SimpleListState.ready(items.where((element) => element.id != 2)),
      ],
      verify: (bloc) => verify(adapter.delete(items[1])).called(1),
    );

    _blocTest(
      'Showld replace the item in the list',
      builder: () {
        when(adapter.update(_TestItem(name: 'Item two', id: 2))).thenAnswer(
            (realInvocation) async => _TestItem(name: 'Item two', id: 2));
        return SimpleListBloc(adapter);
      },
      seed: SimpleListState.ready(items),
      act: (cubit) => cubit.add(
        UpdateItemEvent(
          items[1],
          _TestItem(name: 'Item two', id: 2),
        ),
      ),
      verify: (bloc) {
        expect(
          bloc.state.items,
          isNot(
            contains(items[1]),
          ),
        );
        expect(
          bloc.state.items,
          contains(
            predicate((_TestItem item) => item.id == 2),
          ),
        );
        verify(adapter.update(any)).called(1);
      },
    );

    _blocTest(
      'Showld not add any item if an error occurs',
      builder: () {
        when(adapter.create(any)).thenAnswer((realInvocation) async {
          throw Exception();
        });
        return SimpleListBloc(adapter);
      },
      act: (cubit) => cubit.add(
        CreateItemEvent(
          _TestItem(name: 'named item'),
        ),
      ),
      seed: SimpleListState.ready(items),
      expect: [
        SimpleListState.loading(items),
        predicate(
          (SimpleListState state) =>
              state.loadingState.status == LoadingStatus.error &&
              !state.items.contains(_TestItem(name: 'named item')),
          'Is an error',
        ),
      ],
      verify: (bloc) => verify(adapter.create(any)).called(1),
    );

    _blocTest(
      'Showld not remove any item when an error occurs',
      builder: () {
        when(adapter.delete(items[4])).thenAnswer((realInvocation) async {
          throw Exception();
        });
        return SimpleListBloc(adapter);
      },
      seed: SimpleListState.ready(items),
      act: (cubit) => cubit.add(
        DeleteItemEvent(
          items[4],
        ),
      ),
      expect: [
        SimpleListState.loading(items),
        predicate((SimpleListState state) {
          return state.loadingState.status == LoadingStatus.error &&
              state.items.contains(items[4]);
        }, 'Item is not removed.'),
      ],
      verify: (bloc) => verify(adapter.delete(items[4])).called(1),
    );

    _blocTest(
      'Throws `StateError` when trying to remove a nonexistent item',
      builder: () => SimpleListBloc(adapter),
      act: (cubit) => cubit.add(
        DeleteItemEvent(
          _TestItem(name: 'An item', id: 5),
        ),
      ),
      seed: SimpleListState.ready(items),
      errors: [
        isA<StateError>(),
      ],
      verify: (bloc) => verifyNever(adapter.delete(any)),
    );
    _blocTest(
      'Throws `StateError` when trying to update a nonexistent item',
      builder: () => SimpleListBloc(adapter),
      act: (cubit) => cubit.add(
        UpdateItemEvent(
          _TestItem(name: 'An item', id: 5),
          _TestItem(name: 'An itemm', id: 5),
        ),
      ),
      seed: SimpleListState.ready(items),
      errors: [
        isA<StateError>(),
      ],
      verify: (bloc) => verifyNever(adapter.update(any)),
    );

    _blocTest(
      'Throws `StateError` when trying to retrieve a nonexistent item',
      builder: () => SimpleListBloc(adapter),
      act: (cubit) => cubit.add(
        ReloadItemEvent(
          _TestItem(name: 'An item', id: 5),
        ),
      ),
      seed: SimpleListState.ready(items),
      errors: [
        isA<StateError>(),
      ],
      verify: (bloc) => verifyNever(adapter.retrieve(any)),
    );

    _blocTest(
      'Reloads the passed item`',
      builder: () {
        when(adapter.retrieve(items[1])).thenAnswer(
          (realInvocation) => Future.value(
            _TestItem(name: 'Item two', id: 2),
          ),
        );
        return SimpleListBloc(adapter);
      },
      act: (cubit) => cubit.add(
        ReloadItemEvent(items[1]),
      ),
      seed: SimpleListState.ready(items),
      verify: (bloc) {
        expect(
          bloc.state.items,
          isNot(
            contains(items[1]),
          ),
        );
        expect(bloc.state.loadingState.status, LoadingStatus.ready);
      },
    );
  });
}

@isTest
void _blocTest(
  String description, {
  @required SimpleListBloc<_TestItem> Function() builder,
  SimpleListState<_TestItem> seed,
  Function(SimpleListBloc<_TestItem> cubit) act,
  Iterable errors,
  Iterable expect,
  Function(SimpleListBloc<_TestItem> bloc) verify,
  Duration wait,
  int skip = 0,
}) {
  assert(builder != null);
  blocTest(
    description,
    build: builder,
    act: act,
    errors: errors,
    expect: expect,
    seed: seed,
    skip: skip,
    wait: wait,
    verify: verify,
  );
}

abstract class _TestItem implements Built<_TestItem, _TestItemBuilder> {
  factory _TestItem({
    int id,
    @required String name,
  }) =>
      _$TestItem((b) => b
        ..name = name
        ..id = id);
  _TestItem._();

  @nullable
  int get id;
  String get name;
}

class _TestItemAdapterMock extends Mock implements ListAdapter<_TestItem> {}
